/* Copyright (C) |Meso|Star> 2015-2017 (contact@meso-star.com)
 *
 * This software is a program whose purpose is to test the spp library.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#include "ssp.h"
#include "test_ssp_utils.h"
#include <rsys/clock_time.h>

#define NBS 1000000

int
main(int argc, char** argv)
{
  struct ssp_rng* rng;
  struct mem_allocator allocator;
  struct ssp_ranst_gaussian *gaussian;
  int i;
  double x = 0, x2 = 0;
  double exp_mean = 10, mean;
  double exp_std = 3, std;
  struct time start, end, res;
  char dump[512];
  (void)argc, (void)argv;

  mem_init_proxy_allocator(&allocator, &mem_default_allocator);

  CHECK(ssp_rng_create(&allocator, &ssp_rng_mt19937_64, &rng), RES_OK);

  CHECK(ssp_ranst_gaussian_create(NULL, NULL), RES_BAD_ARG);
  CHECK(ssp_ranst_gaussian_create(NULL, &gaussian), RES_OK);
  CHECK(ssp_ranst_gaussian_ref_put(gaussian), RES_OK);

  CHECK(ssp_ranst_gaussian_create(&allocator, NULL), RES_BAD_ARG);
  CHECK(ssp_ranst_gaussian_create(&allocator, &gaussian), RES_OK);

  CHECK(ssp_ranst_gaussian_setup(NULL, exp_mean, exp_std), RES_BAD_ARG);
  CHECK(ssp_ranst_gaussian_setup(gaussian, exp_mean, -1), RES_BAD_ARG);
  CHECK(ssp_ranst_gaussian_setup(gaussian, exp_mean, exp_std), RES_OK);

  CHECK(ssp_ranst_gaussian_ref_get(NULL), RES_BAD_ARG);
  CHECK(ssp_ranst_gaussian_ref_get(gaussian), RES_OK);

  CHECK(ssp_ranst_gaussian_ref_put(NULL), RES_BAD_ARG);
  CHECK(ssp_ranst_gaussian_ref_put(gaussian), RES_OK);

  time_current(&start);
  FOR_EACH(i, 0, NBS) {
    const double r = ssp_ranst_gaussian_get(gaussian, rng);
    x += r;
    x2 += r * r;
  }
  time_current(&end);
  time_sub(&res, &end, &start);
  time_dump(&res, TIME_SEC|TIME_MSEC|TIME_USEC, NULL, dump, sizeof(dump));
  printf("%s--\n", dump);

  mean = x/NBS;
  std = sqrt(x2/NBS - mean*mean);
  printf("%g %g\n", mean, std);
  CHECK(eq_eps(mean, exp_mean, 1e-2), 1);
  CHECK(eq_eps(std, exp_std, 1e-2), 1);

  /* Same test with inline gaussian generation */
  x = 0;
  x2 = 0;

  time_current(&start);
  FOR_EACH(i, 0, NBS) {
    const double r = ssp_ran_gaussian(rng, exp_mean, exp_std);
    x += r;
    x2 += r * r;
  }
  time_current(&end);
  time_sub(&res, &end, &start);
  time_dump(&res, TIME_SEC|TIME_MSEC|TIME_USEC, NULL, dump, sizeof(dump));
  printf("%s--\n", dump);

  mean = x/NBS;
  std = sqrt(x2/NBS - mean*mean);
  printf("%g %g\n", mean, std);
  CHECK(eq_eps(mean, exp_mean, 1.e-2), 1);
  CHECK(eq_eps(std, exp_std, 1e-2), 1);

  CHECK(ssp_ranst_gaussian_ref_put(gaussian), RES_OK);

  CHECK(ssp_rng_ref_put(rng), RES_OK);

  check_memory_allocator(&allocator);
  mem_shutdown_proxy_allocator(&allocator);
  CHECK(mem_allocated_size(), 0);
  return 0;
}
