/* Copyright (C) |Meso|Star> 2015-2017 (contact@meso-star.com)
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#include "ssp.h"
#include "test_ssp_utils.h"

#include <rsys/double4.h>

#define NSAMPS 128

int
main(int argc, char** argv)
{
  struct ssp_rng* rng0, *rng1;
  struct mem_allocator allocator;
  double samps0[NSAMPS][4];
  double samps1[NSAMPS][4];
  double samps2[NSAMPS][4];
  double samps3[NSAMPS][4];
  int i = 0, j = 0;
  double* f;
  (void)argc, (void)argv;

  mem_init_proxy_allocator(&allocator, &mem_default_allocator);

  CHECK(ssp_rng_create(&allocator, &ssp_rng_mt19937_64, &rng0), RES_OK);
  CHECK(ssp_rng_create(&allocator, &ssp_rng_mt19937_64, &rng1), RES_OK);

  ssp_rng_set(rng0, 0);
  FOR_EACH(i, 0, NSAMPS) {
    double frame[9];
    double up[3] = {0.f, 0.f, 1.f};
    double xyz[3];
    uint64_t seed = ssp_rng_get(rng0);

    ssp_rng_set(rng1, seed);
    f = ssp_ran_hemisphere_uniform_local(rng1, samps0[i]);
    CHECK(f, samps0[i]);
    CHECK(d3_is_normalized(f), 1);
    CHECK(eq_eps(f[3], (float)(1.0/(2.0*PI)), 1.e-6), 1);
    CHECK(eq_eps(f[3], ssp_ran_hemisphere_uniform_local_pdf(f), 1.e-6), 1);
    CHECK(d3_dot(f, up) >= 0.f, 1);

    ssp_rng_set(rng1, seed);
    f = ssp_ran_hemisphere_uniform(rng1, up, samps1[i]);
    CHECK(f, samps1[i]);
    CHECK(d4_eq_eps(f, samps0[i], 1.e-6), 1);

    up[0] = ssp_rng_uniform_double(rng1, -1.0, 1.0);
    up[1] = ssp_rng_uniform_double(rng1, -1.0, 1.0);
    up[2] = ssp_rng_uniform_double(rng1, -1.0, 1.0);
    d3_normalize(up, up);

    ssp_rng_set(rng1, seed);
    f = ssp_ran_hemisphere_uniform(rng1, up, samps1[i]);
    CHECK(d3_eq_eps(samps0[i], samps1[i], 1.e-6), 0);
    CHECK(d3_is_normalized(f), 1);
    CHECK(d3_dot(f, up) >= 0.f, 1);
    CHECK(eq_eps(f[3], 1.0/(2.0*PI), 1.e-6 ), 1);
    CHECK(eq_eps(f[3], ssp_ran_hemisphere_uniform_pdf(up, f), 1.e-6), 1);

    d33_basis(frame, up);
    d33_muld3(xyz, frame, samps0[i]);
    CHECK(d3_eq_eps(samps1[i], xyz, 1.e-6), 1);
    FOR_EACH(j, 0, i) {
      CHECK(d3_eq_eps(samps0[i], samps0[j], 1.e-6), 0);
      CHECK(d3_eq_eps(samps1[i], samps1[j], 1.e-6), 0);
    }
  }

  samps1[1][0] = ssp_rng_uniform_double(rng1, -1.0, 1.0);
  samps1[1][1] = ssp_rng_uniform_double(rng1, -1.0, 1.0);
  samps1[1][2] = ssp_rng_uniform_double(rng1, -1.0, 1.0);
  d3_normalize(samps1[1], samps1[1]);

  ssp_rng_set(rng0, 0);
  ssp_ran_hemisphere_uniform(rng0, samps1[1], samps0[0]);
  ssp_rng_set(rng0, 0);
  ssp_ran_hemisphere_uniform(rng0, samps1[1], samps1[1]);
  CHECK(d4_eq(samps0[0], samps1[1]), 1);

  ssp_rng_set(rng0, 0);
  FOR_EACH(i, 0, NSAMPS) {
    double frame[9];
    double up[3] = { 0.0, 0.0, 1.0 };
    double xyz[3];
    uint64_t seed = ssp_rng_get(rng0);

    ssp_rng_set(rng1, seed);
    f = ssp_ran_hemisphere_cos_local(rng1, samps2[i]);
    CHECK(f, samps2[i]);
    CHECK(d3_eq_eps(samps0[i], samps2[i], 1.e-6), 0);
    CHECK(d3_is_normalized(f), 1);
    CHECK(eq_eps(f[3], f[2]/PI, 1.e-6), 1);
    CHECK(eq_eps(f[3], ssp_ran_hemisphere_cos_local_pdf(f), 1.e-6), 1);
    CHECK(d3_dot(f, up) >= 0.f, 1);

    ssp_rng_set(rng1, seed);
    f = ssp_ran_hemisphere_cos(rng1, up, samps3[i]);
    CHECK(f, samps3[i]);
    CHECK(d4_eq_eps(f, samps2[i], 1.e-6), 1);

    up[0] = ssp_rng_uniform_double(rng1, -1.0, 1.0);
    up[1] = ssp_rng_uniform_double(rng1, -1.0, 1.0);
    up[2] = ssp_rng_uniform_double(rng1, -1.0, 1.0);
    d3_normalize(up, up);

    ssp_rng_set(rng1, seed);
    f = ssp_ran_hemisphere_cos(rng1, up, samps3[i]);
    CHECK(d3_eq_eps(samps2[i], samps3[i], 1.e-6), 0);
    CHECK(d3_eq_eps(samps1[i], samps3[i], 1.e-6), 0);
    CHECK(d3_is_normalized(f), 1);
    CHECK(d3_dot(f, up) >= 0.f, 1);
    CHECK(eq_eps(f[3], d3_dot(f, up)/PI, 1.e-6), 1);
    CHECK(eq_eps(f[3], ssp_ran_hemisphere_cos_pdf(f, up), 1.e-6), 1);

    d33_basis(frame, up);
    d33_muld3(xyz, frame, samps2[i]);
    CHECK(d3_eq_eps(samps3[i], xyz, 1.e-6), 1);
    FOR_EACH(j, 0, i) {
      CHECK(d3_eq_eps(samps2[i], samps2[j], 1.e-6), 0);
      CHECK(d3_eq_eps(samps3[i], samps3[j], 1.e-6), 0);
    }
  }

  samps1[1][0] = ssp_rng_uniform_double(rng1, -1.0, 1.0);
  samps1[1][1] = ssp_rng_uniform_double(rng1, -1.0, 1.0);
  samps1[1][2] = ssp_rng_uniform_double(rng1, -1.0, 1.0);
  d3_normalize(samps1[1], samps1[1]);

  ssp_rng_set(rng0, 0);
  ssp_ran_hemisphere_cos(rng0, samps1[1], samps0[0]);
  ssp_rng_set(rng0, 0);
  ssp_ran_hemisphere_cos(rng0, samps1[1], samps1[1]);
  CHECK(d4_eq(samps0[0], samps1[1]), 1);

  ssp_rng_ref_put(rng0);
  ssp_rng_ref_put(rng1);
  check_memory_allocator(&allocator);
  mem_shutdown_proxy_allocator(&allocator);

  CHECK(mem_allocated_size(), 0);
  return 0;
}
