/* Copyright (C) |Meso|Star> 2015-2017 (contact@meso-star.com)
 *
 * This software is a program whose purpose is to test the spp library.
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#include "ssp.h"
#include "test_ssp_utils.h"

#define NRANDS 2000000
#define COUNT 4605307

static void
test_creation(void)
{
  struct ssp_rng_proxy* proxy;
  struct ssp_rng_type type;

  CHECK(ssp_rng_proxy_create(NULL, NULL, 0, NULL), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_create(NULL, &ssp_rng_mt19937_64, 0, NULL), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_create(NULL, NULL, 4, NULL), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_create(NULL, &ssp_rng_mt19937_64, 4, NULL), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_create(NULL, NULL, 0, &proxy), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_create(NULL, &ssp_rng_mt19937_64, 0, &proxy), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_create(NULL, NULL, 4, &proxy), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_create(NULL, &ssp_rng_mt19937_64, 4, &proxy), RES_OK);

  CHECK(ssp_rng_proxy_get_type(NULL, NULL), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_get_type(proxy, NULL), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_get_type(NULL, &type), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_get_type(proxy, &type), RES_OK);

  CHECK(ssp_rng_type_eq(&type, &ssp_rng_kiss), 0);
  CHECK(ssp_rng_type_eq(&type, &ssp_rng_mt19937_64), 1);

  CHECK(ssp_rng_proxy_ref_get(NULL), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_ref_get(proxy), RES_OK);
  CHECK(ssp_rng_proxy_ref_put(NULL), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_ref_put(proxy), RES_OK);
  CHECK(ssp_rng_proxy_ref_put(proxy), RES_OK);

  CHECK(ssp_rng_proxy_create
    (&mem_default_allocator, &ssp_rng_threefry, 1, &proxy), RES_OK);
  CHECK(ssp_rng_proxy_get_type(proxy, &type), RES_OK);
  CHECK(ssp_rng_type_eq(&type, &ssp_rng_threefry), 1);
  CHECK(ssp_rng_proxy_ref_put(proxy), RES_OK);
}

static void
test_rng_from_proxy(void)
{
  struct ssp_rng_type type;
  struct ssp_rng_proxy* proxy;
  struct ssp_rng* rng[4];
  uint64_t r[4];
  size_t i, j, k;

  CHECK(ssp_rng_proxy_create(NULL, &ssp_rng_mt19937_64, 4, &proxy), RES_OK);

  CHECK(ssp_rng_proxy_create_rng(NULL, 0, NULL), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_create_rng(proxy, 0, NULL), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_create_rng(NULL, 0, &rng[0]), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_create_rng(proxy, 0, &rng[0]), RES_OK);
  CHECK(ssp_rng_proxy_create_rng(proxy, 0, &rng[1]), RES_BAD_ARG);
  CHECK(ssp_rng_ref_put(rng[0]), RES_OK);

  CHECK(ssp_rng_proxy_create_rng(proxy, 4, &rng[0]), RES_BAD_ARG);

  FOR_EACH(i, 0, 4) CHECK(ssp_rng_proxy_create_rng(proxy, i, &rng[i]), RES_OK);
  CHECK(ssp_rng_proxy_ref_put(proxy), RES_OK);

  CHECK(ssp_rng_get_type(rng[0], &type), RES_OK);
  FOR_EACH(i, 1, 4) {
    struct ssp_rng_type type2;
    CHECK(ssp_rng_get_type(rng[i], &type2), RES_OK);
    CHECK(ssp_rng_type_eq(&type, &type2), 1);
  }

  FOR_EACH(i, 0, NRANDS) {
    FOR_EACH(j, 0, 4) {
      r[j] = ssp_rng_get(rng[j]);
      FOR_EACH(k, 0, j) NCHECK(r[k], r[j]);
    }
  }

  FOR_EACH(i, 0, 4) CHECK(ssp_rng_ref_put(rng[i]), RES_OK);
}

static void
test_multi_proxies(void)
{
  struct ssp_rng_proxy* proxy1;
  struct ssp_rng_proxy* proxy2;
  struct ssp_rng* rng1;
  struct ssp_rng* rng2;
  struct ssp_rng* rng3;
  size_t i;

  CHECK(ssp_rng_proxy_create(NULL, &ssp_rng_threefry, 1, &proxy1), RES_OK);
  CHECK(ssp_rng_proxy_create(NULL, &ssp_rng_threefry, 1, &proxy2), RES_OK);

  CHECK(ssp_rng_proxy_create_rng(proxy1, 0, &rng1), RES_OK);
  CHECK(ssp_rng_proxy_create_rng(proxy2, 0, &rng2), RES_OK);
  ssp_rng_create(NULL, &ssp_rng_threefry, &rng3);

  CHECK(ssp_rng_proxy_ref_put(proxy1), RES_OK);
  CHECK(ssp_rng_proxy_ref_put(proxy2), RES_OK);

  ssp_rng_discard(rng1, COUNT);
  ssp_rng_discard(rng3, COUNT+1);
  FOR_EACH(i, 0, COUNT) ssp_rng_get(rng2);

  CHECK(ssp_rng_get(rng1), ssp_rng_get(rng2));
  CHECK(ssp_rng_get(rng1), ssp_rng_get(rng3));

  CHECK(ssp_rng_ref_put(rng1), RES_OK);
  CHECK(ssp_rng_ref_put(rng2), RES_OK);
  CHECK(ssp_rng_ref_put(rng3), RES_OK);
}

static void
test_proxy_from_rng(void)
{
  struct ssp_rng_proxy* proxy;
  struct ssp_rng* rng[4];
  struct ssp_rng_type type;
  uint64_t r[4];
  size_t i, j, k;

  CHECK(ssp_rng_create(NULL, &ssp_rng_threefry, &rng[0]), RES_OK);
  CHECK(ssp_rng_discard(rng[0], COUNT), RES_OK);

  CHECK(ssp_rng_proxy_create_from_rng(NULL, NULL, 0, NULL), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_create_from_rng(NULL, rng[0], 0, NULL), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_create_from_rng(NULL, NULL, 1, NULL), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_create_from_rng(NULL, rng[0], 1, NULL), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_create_from_rng(NULL, NULL, 0, &proxy), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_create_from_rng(NULL, rng[0], 0, &proxy), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_create_from_rng(NULL, NULL, 1, &proxy), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_create_from_rng(NULL, rng[0], 1, &proxy), RES_OK);
  CHECK(ssp_rng_proxy_create_rng(proxy, 0, &rng[1]), RES_OK);
  CHECK(ssp_rng_proxy_create_rng(proxy, 0, &rng[2]), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_create_rng(proxy, 1, &rng[2]), RES_BAD_ARG);

  CHECK(ssp_rng_get(rng[0]), ssp_rng_get(rng[1]));

  CHECK(ssp_rng_ref_put(rng[1]), RES_OK);
  CHECK(ssp_rng_proxy_ref_put(proxy), RES_OK);

  CHECK(ssp_rng_proxy_create_from_rng(NULL, rng[0], 4, &proxy), RES_OK);
  CHECK(ssp_rng_ref_put(rng[0]), RES_OK);
  FOR_EACH(i, 0, 4) CHECK(ssp_rng_proxy_create_rng(proxy, i, &rng[i]), RES_OK);
  CHECK(ssp_rng_proxy_ref_put(proxy), RES_OK);

  CHECK(ssp_rng_get_type(rng[0], &type), RES_OK);
  FOR_EACH(i, 1, 4) {
    struct ssp_rng_type type2;
    CHECK(ssp_rng_get_type(rng[i], &type2), RES_OK);
    CHECK(ssp_rng_type_eq(&type, &type2), 1);
  }

  FOR_EACH(i, 0, NRANDS) {
    FOR_EACH(j, 0, 4) {
      r[j] = ssp_rng_get(rng[j]);
      FOR_EACH(k, 0, j) NCHECK(r[k], r[j]);
    }
  }
  FOR_EACH(i, 0, 4) CHECK(ssp_rng_ref_put(rng[i]), RES_OK);
}

static void
test_read(void)
{
  struct ssp_rng_proxy* proxy;
  struct ssp_rng* rng;
  struct ssp_rng* rng1[4];
  FILE* stream;
  uint64_t r[4];
  size_t i, j, k;

  CHECK(ssp_rng_create(NULL, &ssp_rng_mt19937_64, &rng), RES_OK);
  CHECK(ssp_rng_discard(rng, COUNT), RES_OK);

  stream = tmpfile();
  NCHECK(stream, NULL);
  CHECK(ssp_rng_write(rng, stream), RES_OK);
  rewind(stream);

  CHECK(ssp_rng_proxy_create(NULL, &ssp_rng_mt19937_64, 4, &proxy), RES_OK);
  CHECK(ssp_rng_proxy_read(NULL, NULL), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_read(proxy, NULL), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_read(NULL, stream), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_read(proxy, stream), RES_OK);

  FOR_EACH(i, 0, 4) CHECK(ssp_rng_proxy_create_rng(proxy, i, &rng1[i]), RES_OK);

  r[0] = ssp_rng_get(rng);
  CHECK(r[0],  ssp_rng_get(rng1[0]));
  NCHECK(r[0], ssp_rng_get(rng1[1]));
  NCHECK(r[0], ssp_rng_get(rng1[2]));
  NCHECK(r[0], ssp_rng_get(rng1[3]));

  FOR_EACH(i, 0, NRANDS) {
    FOR_EACH(j, 0, 4) {
      r[j] = ssp_rng_get(rng1[j]);
      FOR_EACH(k, 0, j) NCHECK(r[k], r[j]);
    }
  }

  CHECK(ssp_rng_proxy_ref_put(proxy), RES_OK);
  CHECK(ssp_rng_ref_put(rng), RES_OK);
  FOR_EACH(i, 0, 4) CHECK(ssp_rng_ref_put(rng1[i]), RES_OK);
  fclose(stream);
}

static void
test_write(void)
{
  struct ssp_rng_proxy* proxy;
  struct ssp_rng* rng0;
  struct ssp_rng* rng1;
  FILE* stream;
  uint64_t r;

  stream = tmpfile();
  NCHECK(stream, NULL);

  CHECK(ssp_rng_proxy_create(NULL, &ssp_rng_ranlux48, 1, &proxy), RES_OK);
  CHECK(ssp_rng_proxy_create_rng(proxy, 0, &rng0), RES_OK);

  CHECK(ssp_rng_proxy_write(NULL, NULL), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_write(proxy, NULL), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_write(NULL, stream), RES_BAD_ARG);
  CHECK(ssp_rng_proxy_write(proxy, stream), RES_OK);

  CHECK(ssp_rng_create(NULL, &ssp_rng_ranlux48, &rng1), RES_OK);
  r = ssp_rng_get(rng0);
  CHECK(r, ssp_rng_get(rng1));

  rewind(stream);
  CHECK(ssp_rng_read(rng1, stream), RES_OK);
  CHECK(r, ssp_rng_get(rng1));
  CHECK(ssp_rng_get(rng0), ssp_rng_get(rng1));

  CHECK(ssp_rng_proxy_ref_put(proxy), RES_OK);
  CHECK(ssp_rng_ref_put(rng0), RES_OK);
  CHECK(ssp_rng_ref_put(rng1), RES_OK);

  fclose(stream);
}

int
main(int argc, char** argv)
{
  (void)argc, (void)argv;
  test_creation();
  test_rng_from_proxy();
  test_multi_proxies();
  test_proxy_from_rng();
  test_read();
  test_write();
  CHECK(mem_allocated_size(), 0);
  return 0;
}
